@@ -1,16 +1,17 @@
-from typing import List
+from os.path import commonprefix
+from typing import List, Tuple
 from collections import Counter
 from collections.abc import Mapping
 import numpy as np
 
-from numbers import Number
+from numbers import Number, Integral
 import math
 
 import operator
 import itertools
 from functools import partial
 
-from .struct import struct, Struct, transpose_structs, map_type
+from .struct import struct, Struct, transpose_structs, map_type, traverse_type, reduce_type
 
 
 class Table(Struct):
@@ -18,60 +19,83 @@ class Table(Struct):
 
         assert len(d) > 0, "empty Table"
         t = next(iter(d.values()))
-      
+
         for (k, v) in d.items():
-            assert type(v) == np.ndarray, "expected ndarray, got " + type(t).__name__
-            assert v.shape[0] == t.shape[0], "mismatched column sizes: " + str(shape(d))
+            assert hasattr(v, 'shape'), "expected ndarray, got " + type(t).__name__
+
+        shapes = [t.shape for t in d.values()]
+        self._prefix = commonprefix(shapes)
 
         super(Table, self).__init__(d) 
 
     @staticmethod
     def from_structs(structs : List[Struct]) -> 'Table':
         struct_lists = transpose_structs(structs)
-        return Table(struct_lists._map(np.stack).__dict__)
-   
+        return Table(struct_lists._map(np.stack)._entries)
+
+    @staticmethod
+    def singleton(struct : Struct) -> 'Table':     
+        return Table(struct._map(np.expand_dims, axis=0))
+
+    @property
+    def _shapes(self) -> 'Struct':     
+        return self._map(lambda x: x.shape)
+
+    @property
+    def _shape(self) -> Tuple:
+        return self._prefix
+
     @staticmethod 
-    def build(**d):
+    def stack(structs : List[Struct]) -> 'Table':
+        return Table.from_structs(structs)
+
+    @staticmethod
+    def build(d:dict):
+        is_array = [isinstance(t, np.ndarray) for t in d.values()]
+        if all(is_array):
+            return Table(d)
+        else:
+            return Struct(d)
+
+    @staticmethod
+    def create(**d):
         return Table(d)
 
-    def __getitem__(self, index:int) -> np.ndarray:
-        return self.__dict__[index]        
+    def _index_select(self, index:np.ndarray, axis=0) -> 'Table':
+        assert axis < len(self._prefix)
 
-    def _index_select(self, index:np.ndarray) -> 'Table':
         if isinstance(index, np.ndarray):
-            assert index.dtype == torch.int64 
+            assert issubclass(index.dtype.type, Integral)
             assert index.dim() == 1
         else:
-            assert isinstance(index, int), "Table.index_select: unsupported index type" + type(index).__name__
+            assert issubclass(type(index), Integral),\
+                "Table.index_select: unsupported index type" + type(index).__name__
 
-        return self._map(lambda t: t[index])
-
-
-    def _index(self, index:int) -> np.ndarray:
-        return self._index_select(np.ndarray([index], dtype=int))
+        return self._map(lambda t: np.take(t, index, axis=axis))
 
         
-    def _narrow(self, start, n):
-        return self._map(lambda t: t.narrow(0, start, n))
+    def _narrow(self, start, n, axis=0):
+        return self._index_select(np.arange(start, start + n), axis=axis)
 
 
-    def _take(self, n):
-        return self._narrow(0, min(self._size, n))
+    def _take(self, n, axis=0): 
+        return self._narrow(0, min(self._prefix[axis], n), axis=axis)
 
-    def _drop(self, n):
-        n = min(self._size, n)
-        return self._narrow(n, self._size - n)
+    def _drop(self, n, axis=0):
+        n = min(self._prefix[axis], n)
+        return self._narrow(n, self._prefix[axis] - n, axis=axis)
 
 
-    def _sequence(self):
-        return (self._index_select(i) for i in range(0, self._size))
+    def _sequence(self, axis = 0):
+        return (self._index_select(i, axis=axis) for i in range(0, self._prefix[axis]))
 
-    def _sort_on(self, key, descending=False):
+    def _sort_on(self, key, descending=False, axis=0):
         assert key in self
         assert self[key].dim() == 1
 
-        values, inds = self[key].sort(descending = descending)
-        return Table({k: values if k == key else v[inds] for k, v in self.items()})
+        values, inds = self[key].sort(descending = descending, axis=axis)
+        return Table({k: values if k == key else np.take(v, inds, axis=axis) 
+            for k, v in self.items()})
 
     @property
     def _head(self):
@@ -81,24 +105,11 @@ class Table(Struct):
     def _size(self):
         return self._head.shape[0]
 
-    @property
-    def _device(self):
-        return self._head.device
-
-    def _to(self, device):
-        return self._map(lambda t: t.to(device))
-
-    def _cpu(self):
-        return self._map(lambda t: t.cpu())
-
-
 
 def table(**d):
     return Table(d)
 
 
-
-
 class Histogram:
     def __init__(self, values = np.array([0]), range = (0, 1), num_bins = 10, trim = True):
         assert len(range) == 2
@@ -178,6 +189,14 @@ class Histogram:
 def map_arrays(data, f, *args, **kwargs):
     return map_type(data,  np.ndarray, partial(f, *args, **kwargs))  
 
+def traverse_arrays(data, f, *args, **kwargs):
+    return traverse_type(data,  np.ndarray, partial(f, *args, **kwargs))  
+
+def reduce_arrays(data, f, op, initializer=None):
+    return reduce_type(data,  np.ndarray, f, op, initializer=initializer)  
+
+
+
 def shape_info(x):
     return map_arrays(x, lambda x: tuple([*x.shape, type(x), x.dtype]))
 
@@ -195,3 +214,5 @@ def stack_tables(tables, dim=0):
 def cat_tables(tables, dim=0):
     t = transpose_structs(tables)
     return Table(dict(t._map(np.concatenate, axis=dim))) 
+
+
